
/*  XMMS - Cross-platform multimedia player
 *  Copyright (C) 1998-1999  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include "cdaudio.h"

InputPlugin cdda_ip =
{
	NULL,
	NULL,
	"CD Audio Player 0.9",
	cdda_init,
	NULL,
	cdda_configure,
	is_our_file,
	scan_dir,
	play_file,
	stop,
	cdda_pause,
	seek,
	NULL,
	get_time,
	get_volume,
	set_volume,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	get_song_info,
	NULL,			/* file_info_box */
	NULL
};

CDDAConfig cdda_cfg;

static gint cdda_fd = -1;
static gint track, ntrack, start_frame;
static gint length;
static gboolean is_paused;

static gint volume_left = 100, volume_right = 100;

InputPlugin *get_iplugin_info(void)
{
	return &cdda_ip;
}

static void cdda_init(void)
{
	ConfigFile *cfgfile;
	gchar *filename;

	filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL);

	cdda_cfg.device = NULL;
	cdda_cfg.directory = NULL;

#if defined(HAVE_SYS_SOUNDCARD_H) || defined(HAVE_MACHINE_SOUNDCARD_H)
	cdda_cfg.use_oss_mixer = TRUE;
#endif

	if (cfgfile = xmms_cfg_open_file(filename))
	{
		xmms_cfg_read_string(cfgfile, "CDDA", "device", &cdda_cfg.device);
		xmms_cfg_read_string(cfgfile, "CDDA", "directory", &cdda_cfg.directory);
		xmms_cfg_read_boolean(cfgfile, "CDDA", "use_oss_mixer", &cdda_cfg.use_oss_mixer);
		xmms_cfg_free(cfgfile);
	}
	g_free(filename);
#ifdef HAVE_SYS_CDIO_H
# ifdef __FreeBSD__
	if (!cdda_cfg.device)
		cdda_cfg.device = g_strdup("/dev/acd0c");
	if (!cdda_cfg.directory)
		cdda_cfg.directory = g_strdup("/cdrom");
# else
	if (!cdda_cfg.device)
		cdda_cfg.device = g_strdup("/vol/dev/aliases/cdrom0");
	if (!cdda_cfg.directory)
		cdda_cfg.directory = g_strdup("/cdrom/cdrom0");
# endif
#else
	if (!cdda_cfg.device)
		cdda_cfg.device = g_strdup("/dev/cdrom");
	if (!cdda_cfg.directory)
		cdda_cfg.directory = g_strdup("/mnt/cdrom/");
#endif

}

static int is_our_file(char *filename)
{
	gchar *ext;

	ext = strrchr(filename, '.');
	if (ext)
		if (!strcasecmp(ext, ".cda"))
			return TRUE;

	return FALSE;
}

static gboolean is_mounted(gchar * device_name)
{
#ifdef HAVE_MNTENT_H
	FILE *mounts;
	struct mntent *mnt;
	char devname[256];
	struct stat st;

	if (lstat(device_name, &st) < 0)
		return -1;

	if (S_ISLNK(st.st_mode))
		readlink(device_name, devname, 256);
	else
		strncpy(devname, device_name, 256);

	if ((mounts = setmntent(MOUNTED, "r")) == NULL)
		return TRUE;

	while ((mnt = getmntent(mounts)) != NULL)
	{
		if (strcmp(mnt->mnt_fsname, devname) == 0)
		{
			endmntent(mounts);
			return TRUE;
		}
	}
	endmntent(mounts);
#endif
	return FALSE;
}

static GList *scan_dir(char *dir)
{
	GList *list = NULL;
	int fd, i;
#if defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__)
	struct ioc_toc_header tochdr;
	struct ioc_read_toc_single_entry tocentry;
#else	
	struct cdrom_tochdr tochdr;
	struct cdrom_tocentry tocentry;
#endif

	if (is_mounted(cdda_cfg.device))
		return NULL;

	if (strcmp(dir, cdda_cfg.directory))
		return NULL;

	fd = open(cdda_cfg.device, CDOPENFLAGS);
	if (fd == -1)
		return NULL;

#if defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__)
	if (!ioctl(fd, CDIOREADTOCHEADER, &tochdr))
	{
		for (i = tochdr.starting_track; i <= tochdr.ending_track; i++)
		{
			tocentry.track = i;
			tocentry.address_format = CD_MSF_FORMAT;
			if (!ioctl(fd, CDIOREADTOCENTRY, &tocentry))
			{
				if (~tocentry.entry.control & 4)
	list = g_list_append(list, g_strdup_printf("Track %02d.cda", i));
			}
		}
	}
#else	
	if (!ioctl(fd, CDROMREADTOCHDR, &tochdr))
	{
		for (i = tochdr.cdth_trk0; i <= tochdr.cdth_trk1; i++)
		{
			tocentry.cdte_track = i;
			tocentry.cdte_format = CDROM_MSF;
			if (!ioctl(fd, CDROMREADTOCENTRY, &tocentry))
			{
				if (tocentry.cdte_ctrl != CDROM_DATA_TRACK)
					list = g_list_append(list, g_strdup_printf("Track %02d.cda", i));
			}
		}
	}
#endif
	close(fd);

	return list;
}

static void play_file(char *filename)
{
	gchar *tmp;
#if defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__)
	struct ioc_read_toc_single_entry tocentry;
	struct ioc_toc_header tochdr;
#else
	struct cdrom_tocentry tocentry;
	struct cdrom_tochdr tochdr;
#endif

	if (is_mounted(cdda_cfg.device))
		return;

	tmp = strrchr(filename, '/');
	if (tmp)
		tmp++;
	else
		tmp = filename;

	if (!sscanf(tmp, "Track %d.cda", &track))
		return;

	cdda_fd = open(cdda_cfg.device, CDOPENFLAGS);
	if (cdda_fd == -1)
		return;

#if defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__)
	ioctl(cdda_fd, CDIOREADTOCHEADER, &tochdr);
	if (track < tochdr.starting_track || track > tochdr.ending_track)
	{
		close(cdda_fd);
		cdda_fd = -1;
		return;
	}
	ntrack = track + 1;

	tocentry.track = ntrack;
	tocentry.address_format = CD_MSF_FORMAT;
	ioctl(cdda_fd, CDIOREADTOCENTRY, &tocentry);
	length = (tocentry.entry.addr.msf.minute * 60 +
		  tocentry.entry.addr.msf.second) * CD_FRAMES +
	    tocentry.entry.addr.msf.frame;

	tocentry.track = track;
	tocentry.address_format = CD_MSF_FORMAT;
	ioctl(cdda_fd, CDIOREADTOCENTRY, &tocentry);
	length -= (tocentry.entry.addr.msf.minute * 60 +
		   tocentry.entry.addr.msf.second) * CD_FRAMES +
	    tocentry.entry.addr.msf.frame;
	start_frame = (tocentry.entry.addr.msf.minute * 60 +
		       tocentry.entry.addr.msf.second) * CD_FRAMES +
	    tocentry.entry.addr.msf.frame;
	if (tocentry.entry.control & 4)
	{
		close(cdda_fd);
		cdda_fd = -1;
		return;
	}
#else
	ioctl(cdda_fd, CDROMREADTOCHDR, &tochdr);
	if (track < tochdr.cdth_trk0 || track > tochdr.cdth_trk1)
	{
		close(cdda_fd);
		cdda_fd = -1;
		return;
	}
	ntrack = (tochdr.cdth_trk1 == track) ? CDROM_LEADOUT : track + 1;

	tocentry.cdte_track = ntrack;
	tocentry.cdte_format = CDROM_MSF;
	ioctl(cdda_fd, CDROMREADTOCENTRY, &tocentry);
	length = (tocentry.cdte_addr.msf.minute * 60 +
		  tocentry.cdte_addr.msf.second) * CD_FRAMES +
		tocentry.cdte_addr.msf.frame;

	tocentry.cdte_track = track;
	tocentry.cdte_format = CDROM_MSF;
	ioctl(cdda_fd, CDROMREADTOCENTRY, &tocentry);
	length -= (tocentry.cdte_addr.msf.minute * 60 +
		   tocentry.cdte_addr.msf.second) * CD_FRAMES +
		tocentry.cdte_addr.msf.frame;
	start_frame = (tocentry.cdte_addr.msf.minute * 60 +
		       tocentry.cdte_addr.msf.second) * CD_FRAMES +
		tocentry.cdte_addr.msf.frame;
	if (tocentry.cdte_ctrl == CDROM_DATA_TRACK)
	{
		close(cdda_fd);
		cdda_fd = -1;
		return;
	}
#endif

	tmp = g_strdup_printf("CD Audio Track %02u", track);
	cdda_ip.set_info(tmp, (length * 1000) / 75, 44100 * 2 * 2 * 8,
			 44100, 2);

	is_paused = FALSE;

	seek(0);
}

static void stop(void)
{
#if defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__)
	ioctl(cdda_fd, CDIOCSTOP, 0);
#else
	ioctl(cdda_fd, CDROMSTOP, 0);
#endif
	close(cdda_fd);
	cdda_fd = -1;
}

static void cdda_pause(short p)
{
#if defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__)
	ioctl(cdda_fd, p ? CDIOCPAUSE : CDIOCRESUME);
#else
	ioctl(cdda_fd, p ? CDROMPAUSE : CDROMRESUME);
#endif
	is_paused = (p ? TRUE : FALSE);
}

static void seek(int time)
#if defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__)
{
	struct ioc_read_toc_single_entry tocentry;
	struct ioc_play_msf msf;

	tocentry.track = track;
	tocentry.address_format = CD_MSF_FORMAT;
	ioctl(cdda_fd, CDIOREADTOCENTRY, &tocentry);
	time += tocentry.entry.addr.msf.minute * 60 +
	    tocentry.entry.addr.msf.second;
	msf.start_m = time / 60;
	msf.start_s = time % 60;
	msf.start_f = 0;
	tocentry.track = ntrack;
	ioctl(cdda_fd, CDIOREADTOCENTRY, &tocentry);
	msf.end_m = tocentry.entry.addr.msf.minute;
	msf.end_s = tocentry.entry.addr.msf.second;
	msf.end_f = tocentry.entry.addr.msf.frame;
	ioctl(cdda_fd, CDIOCPLAYMSF, &msf);
	if (is_paused)
		ioctl(cdda_fd, CDIOCPAUSE);
}
#else
{
	struct cdrom_tocentry tocentry;
	struct cdrom_msf msf;

	tocentry.cdte_track = track;
	tocentry.cdte_format = CDROM_MSF;
	ioctl(cdda_fd, CDROMREADTOCENTRY, &tocentry);
	time += tocentry.cdte_addr.msf.minute * 60 + tocentry.cdte_addr.msf.second;
	msf.cdmsf_min0 = time / 60;
	msf.cdmsf_sec0 = time % 60;
	msf.cdmsf_frame0 = 0;
	tocentry.cdte_track = ntrack;
	ioctl(cdda_fd, CDROMREADTOCENTRY, &tocentry);
	msf.cdmsf_min1 = tocentry.cdte_addr.msf.minute;
	msf.cdmsf_sec1 = tocentry.cdte_addr.msf.second;
	msf.cdmsf_frame1 = tocentry.cdte_addr.msf.frame;
	ioctl(cdda_fd, CDROMPLAYMSF, &msf);
	if (is_paused)
		ioctl(cdda_fd, CDROMPAUSE);
}
#endif

static int get_time(void)
#if defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__)
{
	struct ioc_read_subchannel subchnl;
	struct cd_sub_channel_info subinfo;
	gint frame;

	if (cdda_fd == -1)
		return -1;

	subchnl.address_format = CD_MSF_FORMAT;
	subchnl.data_format = CD_CURRENT_POSITION;
	subchnl.data_len = sizeof(subinfo);
	subchnl.data = &subinfo;
	ioctl(cdda_fd, CDIOCREADSUBCHANNEL, &subchnl);

	frame = (subchnl.data->what.position.absaddr.msf.minute * 60 +
		 subchnl.data->what.position.absaddr.msf.second) * CD_FRAMES +
	    subchnl.data->what.position.absaddr.msf.frame;

	if (frame - start_frame >= length - 20)	/* 20 seems to work better */
	{
		stop();
		return -1;
	}

	return ((frame - start_frame) * 1000) / 75;
}
#else
{
	struct cdrom_subchnl subchnl;
	gint frame;

	if (cdda_fd == -1)
		return -1;

	subchnl.cdsc_format = CDROM_MSF;
	ioctl(cdda_fd, CDROMSUBCHNL, &subchnl);

	frame = (subchnl.cdsc_absaddr.msf.minute * 60 + subchnl.cdsc_absaddr.msf.second) * CD_FRAMES + subchnl.cdsc_absaddr.msf.frame;

	if (frame - start_frame >= length - 20)		/* 20 seems to work better */
	{
		stop();
		return -1;
	}

	return ((frame - start_frame) * 1000) / 75;
}
#endif

static void get_song_info(char *filename, char **title, int *len)
{
	gint t, fd;
	gchar *tmp;
#if defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__)
	struct ioc_read_toc_single_entry tocentry;
	struct ioc_toc_header tochdr;
#else
	struct cdrom_tocentry tocentry;
	struct cdrom_tochdr tochdr;
	gint nt;
#endif

	*title = NULL;
	*len = -1;

	if (is_mounted(cdda_cfg.device))
		return;

	tmp = strrchr(filename, '/');
	if (tmp)
		tmp++;
	else
		tmp = filename;

	if (!sscanf(tmp, "Track %d.cda", &t))
		return;

	fd = open(cdda_cfg.device, CDOPENFLAGS);
	if (fd == -1)
		return;

#if defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__)
	ioctl(fd, CDIOREADTOCHEADER, &tochdr);
	if (t < tochdr.starting_track || t > tochdr.ending_track)
	{
		close(fd);
		return;
	}

	tocentry.track = t+1;
	tocentry.address_format = CD_MSF_FORMAT;
	ioctl(fd, CDIOREADTOCENTRY, &tocentry);

	*len = (tocentry.entry.addr.msf.minute * 60 +
		tocentry.entry.addr.msf.second) * CD_FRAMES +
		tocentry.entry.addr.msf.frame;

	tocentry.track = t;
	tocentry.address_format = CD_MSF_FORMAT;
	ioctl(fd, CDIOREADTOCENTRY, &tocentry);
	if (tocentry.entry.control & 4)
	{
		*len = -1;
		close(fd);
		return;
	}


	(*len) -= (tocentry.entry.addr.msf.minute * 60 +
		   tocentry.entry.addr.msf.second) * CD_FRAMES +
		tocentry.entry.addr.msf.frame;
#else
	ioctl(fd, CDROMREADTOCHDR, &tochdr);
	if (t < tochdr.cdth_trk0 || t > tochdr.cdth_trk1)
	{
		close(fd);
		return;
	}
	nt = (tochdr.cdth_trk1 == t) ? CDROM_LEADOUT : t + 1;

	tocentry.cdte_track = nt;
	tocentry.cdte_format = CDROM_MSF;
	ioctl(fd, CDROMREADTOCENTRY, &tocentry);

	*len = (tocentry.cdte_addr.msf.minute * 60 +
		tocentry.cdte_addr.msf.second) * CD_FRAMES +
		tocentry.cdte_addr.msf.frame;

	tocentry.cdte_track = t;
	tocentry.cdte_format = CDROM_MSF;
	ioctl(fd, CDROMREADTOCENTRY, &tocentry);
	if (tocentry.cdte_ctrl == CDROM_DATA_TRACK)
	{
		*len = -1;
		close(fd);
		return;
	}

	(*len) -= (tocentry.cdte_addr.msf.minute * 60 +
		   tocentry.cdte_addr.msf.second) * CD_FRAMES +
		tocentry.cdte_addr.msf.frame;
#endif

	*len = (*len * 1000) / 75;

	*title = g_strdup_printf("CD Audio Track %02u", t);
	close(fd);

}

static void get_volume(int *l, int *r)
{
	int fd, devs, cmd, v;

#if defined(HAVE_SYS_SOUNDCARD_H) || defined(HAVE_MACHINE_SOUNDCARD_H)
	if (cdda_cfg.use_oss_mixer)
	{
		fd = open("/dev/mixer", O_RDONLY);
		if (fd != -1)
		{
			ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devs);
			if (devs & SOUND_MASK_CD)
				cmd = SOUND_MIXER_READ_CD;
			else if (devs & SOUND_MASK_VOLUME)
				cmd = SOUND_MIXER_READ_VOLUME;
			else
			{
				close(fd);
				return;
			}
			ioctl(fd, cmd, &v);
			*r = (v & 0xFF00) >> 8;
			*l = (v & 0x00FF);
			close(fd);
		}
	}
	else
#endif
	if (!cdda_cfg.use_oss_mixer)
	{

#if defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__)
		struct ioc_vol vol;

		if (cdda_fd != -1)
		{
			ioctl(cdda_fd, CDIOCGETVOL, &vol);
			*l = (100 * vol.vol[0]) / 255;
			*r = (100 * vol.vol[1]) / 255;
		}
#elif defined(CDROMVOLREAD)
		struct cdrom_volctrl vol;

		if (cdda_fd != -1)
		{
			ioctl(cdda_fd, CDROMVOLREAD, &vol);
			*l = (100 * vol.channel0) / 255;
			*r = (100 * vol.channel1) / 255;
		}
#else
		*l = volume_left;
		*r = volume_right;
#endif

	}
}

static void set_volume(int l, int r)
{
#if defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__)
	struct ioc_vol vol;
#else
	struct cdrom_volctrl vol;
#endif
	int fd, devs, cmd, v;

#if defined(HAVE_SYS_SOUNDCARD_H) || defined(HAVE_MACHINE_SOUNDCARD_H)
	if (cdda_cfg.use_oss_mixer)
	{
		fd = open("/dev/mixer", O_RDONLY);
		if (fd != -1)
		{
			ioctl(fd, SOUND_MIXER_READ_DEVMASK, &devs);
			if (devs & SOUND_MASK_CD)
				cmd = SOUND_MIXER_WRITE_CD;
			else if (devs & SOUND_MASK_VOLUME)
				cmd = SOUND_MIXER_WRITE_VOLUME;
			else
			{
				close(fd);
				return;
			}
			v = (r << 8) | l;
			ioctl(fd, cmd, &v);
			close(fd);
		}
	}
	else
#endif
	{
		if (cdda_fd != -1)
		{
#if defined(HAVE_SYS_CDIO_H) && defined(__FreeBSD__)
			vol.vol[0] = vol.vol[2] = (l * 255) / 100;
			vol.vol[1] = vol.vol[3] = (r * 255) / 100;
			ioctl(cdda_fd, CDIOCSETVOL, &vol);
#else
			vol.channel0 = vol.channel2 = (l * 255) / 100;
			vol.channel1 = vol.channel3 = (r * 255) / 100;
			ioctl(cdda_fd, CDROMVOLCTRL, &vol);
#endif
		}
#if !defined(CDROMVOLREAD) && !defined(HAVE_SYS_CDIO_H)
		volume_left = l;
		volume_right = r;
#endif
	}
}
